# -*- coding: utf-8 -*-

import os
import tempfile
from collections import defaultdict
from datetime import datetime

import keras.models
import numpy as np
from keras.applications import mobilenet
from keras.utils.generic_utils import CustomObjectScope


def ensure_dir(directory):
    """Create diretory if it does not exist yet."""
    if not os.path.exists(directory):
        os.makedirs(directory)


def make_dumpable(params, datetimes=False):
    """Get nested dicts of params to allow json dumping.

    Also work around this: https://github.com/HDI-Project/BTB/issues/79
    And convert numpy types to primitive types.

    Optionally dump datetimes to ISO format.

    Args:
        params (dict):
            Params dictionary with tuples as keys.
        datetimes (bool):
            whether to convert datetimes to ISO strings or not.

    Returns:
        dict:
            Dumpable params as a tree of dicts and nested sub-dicts.
    """
    nested_params = defaultdict(dict)
    for (block, param), value in params.items():
        if isinstance(value, np.integer):
            value = int(value)

        elif isinstance(value, np.floating):
            value = float(value)

        elif isinstance(value, np.ndarray):
            value = value.tolist()

        elif isinstance(value, np.bool_):
            value = bool(value)

        elif value == 'None':
            value = None

        elif datetimes and isinstance(value, datetime):
            value = value.isoformat()

        nested_params[block][param] = value

    return nested_params


def _walk(document, transform):
    if not isinstance(document, dict):
        return document

    new_doc = dict()
    for key, value in document.items():
        if isinstance(value, dict):
            value = _walk(value, transform)
        elif isinstance(value, list):
            value = [_walk(v, transform) for v in value]

        new_key, new_value = transform(key, value)
        new_doc[new_key] = new_value

    return new_doc


def remove_dots(document):
    """Replace dots with dashes in all the keys from the dictionary."""
    return _walk(document, lambda key, value: (key.replace('.', '-'), value))


def restore_dots(document):
    """Replace dashes with dots in all the keys from the dictionary."""
    return _walk(document, lambda key, value: (key.replace('-', '.'), value))


def make_keras_picklable():
    """Make the keras models picklable."""
    def __getstate__(self):
        model_str = ""
        with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=True) as fd:
            keras.models.save_model(self, fd.name, overwrite=True)
            model_str = fd.read()

        return {'model_str': model_str}

    def __setstate__(self, state):
        with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=True) as fd:
            fd.write(state['model_str'])
            fd.flush()
            try:
                model = keras.models.load_model(fd.name)

            except ValueError:
                scope = {
                    'relu6': mobilenet.relu6,
                    'DepthwiseConv2D': mobilenet.DepthwiseConv2D
                }
                with CustomObjectScope(scope):
                    model = keras.models.load_model(fd.name)

        self.__dict__ = model.__dict__

    cls = keras.models.Model
    cls.__getstate__ = __getstate__
    cls.__setstate__ = __setstate__



# -*- coding: utf-8 -*-

"""AutoBazaar Pipeline Module."""

import json
import logging
import os
import random
import uuid
from collections import Counter

import cloudpickle
import numpy as np
import pandas as pd
from mit_d3m.loaders import get_loader
from mit_d3m.metrics import METRICS_DICT
from mlblocks import MLPipeline

LOGGER = logging.getLogger(__name__)


class ABPipeline(object):
    """AutoBazaar Pipeline Class."""

    def _extract_hyperparameters(self, preprocessing_primitives):
        block_names_count = Counter()
        block_names = list()
        for primitive in preprocessing_primitives:
            block_names_count.update([primitive])
            block_count = block_names_count[primitive]
            block_names.append('{}#{}'.format(primitive, block_count))

        pre_params = dict()
        hyperparameters = self.pipeline_dict['hyperparameters'].copy()
        for block_name in block_names:
            block_params = hyperparameters.pop(block_name, None)
            if block_params:
                pre_params[block_name] = block_params

        return pre_params, hyperparameters

    def __init__(self, pipeline_dict, loader, metric, problem_doc):
        self.pipeline_dict = pipeline_dict
        self.name = pipeline_dict['name']
        self.template = pipeline_dict.get('template')
        self.loader = loader
        self.metric = metric
        self.problem_doc = problem_doc

        preprocessing_blocks = self.pipeline_dict.get('preprocessing_blocks')
        if preprocessing_blocks:
            preprocessing = pipeline_dict.copy()
            preprocessing_primitives = preprocessing['primitives'][:preprocessing_blocks]
            preprocessing['primitives'] = preprocessing_primitives
            self._preprocessing = preprocessing

            tunable = pipeline_dict.copy()
            tunable_primitives = tunable['primitives'][preprocessing_blocks:]
            tunable['primitives'] = tunable_primitives
            self._tunable = tunable

            pre_params, tun_params = self._extract_hyperparameters(preprocessing_primitives)
            self._preprocessing['hyperparameters'] = pre_params
            self._tunable['hyperparameters'] = tun_params

        else:
            self._preprocessing = None
            self._tunable = pipeline_dict

        self.id = str(uuid.uuid4())
        self.cv_scores = list()

        self.rank = None
        self.score = None
        self.dumped = False
        self.fitted = False

        self.pipeline = MLPipeline.from_dict(pipeline_dict)

    def fit(self, data_params):
        """Fit the pipeline on the given params."""
        X, y = data_params.X, data_params.y

        self.pipeline = MLPipeline.from_dict(self.pipeline_dict)
        self.pipeline.fit(X, y, **data_params.context)

        self.fitted = True

    def predict(self, d3mds):
        """Get predictions for the given D3MDS."""
        data_params = self.loader.load(d3mds)

        predictions = self.pipeline.predict(data_params.X, **data_params.context)

        out_df = pd.DataFrame()
        out_df['d3mIndex'] = data_params.y.index
        out_df[d3mds.target_column] = predictions

        return out_df

    def _get_split(self, X, y, indexes):
        if hasattr(X, 'iloc'):
            X = X.iloc[indexes]
        else:
            X = X[indexes]

        if y is not None:
            if hasattr(y, 'iloc'):
                y = y.iloc[indexes]
            else:
                y = y[indexes]

        return X, y

    def _get_score(self):
        score = np.mean(self.cv_scores)
        std = np.std(self.cv_scores)

        if 'Error' in self.metric:
            rank = score

        elif score <= 1:
            rank = 1 - score

        else:
            raise ValueError("Found a score > 1 in a maximization problem: {}".format(score))

        return score, std, rank

    def preprocess(self, X, y, context):
        """Execute the preprocessing steps of the pipeline."""
        if self._preprocessing:
            LOGGER.info("Executing preprocessing pipeline")
            pipeline = MLPipeline.from_dict(self._preprocessing)
            pipeline.fit(X, y, **context)
            return pipeline.predict(X, **context)
        else:
            LOGGER.info("No preprocessing steps found")
            return X

    def cv_score(self, X, y, context, metric=None, cv=None):
        """Cross Validate this pipeline."""

        scorer = METRICS_DICT[metric or self.metric]

        LOGGER.debug('CV Scoring pipeline %s', self)

        self.cv_scores = list()

        for fold, (train_index, test_index) in enumerate(cv.split(X, y)):

            LOGGER.debug('Scoring fold: %s', fold)

            X_train, y_train = self._get_split(X, y, train_index)
            pipeline = MLPipeline.from_dict(self._tunable)
            pipeline.fit(X_train, y_train, **context)

            X_test, y_test = self._get_split(X, y, test_index)
            pred = pipeline.predict(X_test, **context)
            score = scorer(pred, y_test)
            self.cv_scores.append(score)

            LOGGER.debug('Fold %s score: %s', fold, score)

        score, std, rank = self._get_score()

        LOGGER.debug('CV score: %s +/- %s; rank: %s', score, std, rank)

        self.score = score
        self.std = std
        self.rank = rank + random.random() * 1.e-12   # to avoid collisions

    def to_dict(self, problem_doc=False):
        """Return the details of this pipeline in a dict."""
        pipeline_dict = self.pipeline.to_dict().copy()
        pipeline_dict.update({
            'id': self.id,
            'name': self.name,
            'template': self.template,
            'loader': self.loader.to_dict(),
            'score': self.score,
            'rank': self.rank,
            'metric': self.metric
        })
        if problem_doc:
            pipeline_dict['problem_doc'] = self.problem_doc

        return pipeline_dict

    def __repr__(self):
        return 'ABPipeline({})'.format(json.dumps(self.to_dict(), indent=4))

    @classmethod
    def from_dict(cls, pipeline_dict):
        """Load a pipeline from a dict."""
        pipeline_dict = pipeline_dict.copy()
        loader = get_loader(**pipeline_dict.pop('loader'))
        metric = pipeline_dict['metric']
        problem_doc = pipeline_dict.pop('problem_doc')
        return cls(pipeline_dict, loader, metric, problem_doc)

    def dump(self, output_dir, rank=None):
        """Dump this pipeline using pickle."""
        if rank is None:
            rank = self.rank

        LOGGER.info('Dumping pipeline with rank %s: %s', rank, self.id)

        self.dumped = True

        pickle_path = os.path.join(output_dir, '{}.pkl'.format(self.id))
        with open(pickle_path, "wb") as pickle_file:
            LOGGER.info("Outputting pipeline %s", pickle_file.name)
            cloudpickle.dump(self, pickle_file)

        json_path = os.path.join(output_dir, '{}.json'.format(self.id))
        self.pipeline.save(json_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""AutoBazaar Command Line Module."""

import argparse
import json
import os
import shutil
import sys
import traceback
import warnings
from datetime import datetime

import cloudpickle
import pandas as pd
from mit_d3m import metrics
from mit_d3m.dataset import D3MDS
from mit_d3m.db import get_db
from mit_d3m.stats import get_stats
from mit_d3m.utils import logging_setup, make_abs

from autobazaar.search import TUNERS, PipelineSearcher
from autobazaar.utils import make_keras_picklable

warnings.filterwarnings(action='ignore')


def _load_targets(datasets_dir, dataset, problem):
    score_phase = 'SCORE'
    if problem:
        score_phase += '_' + problem

    score_dir = os.path.join(datasets_dir, dataset, score_phase)

    return pd.read_csv(os.path.join(score_dir, 'targets.csv'), index_col='d3mIndex')


def _get_metric(problem_path):
    problem_schema = os.path.join(problem_path, 'problemDoc.json')

    with open(problem_schema, 'r') as f:
        problem_doc = json.load(f)

    problem_metrics = problem_doc['inputs']['performanceMetrics']
    if len(problem_metrics) > 1:
        raise Exception("Wrong number of metrics")

    return metrics.METRICS_DICT[problem_metrics[0]['metric']]


def _get_dataset_paths(datasets_dir, dataset, phase, problem):
    if problem:
        full_phase = phase + '_' + problem
    else:
        full_phase = phase

    root_dir = os.path.join(datasets_dir, dataset, full_phase)
    dataset_path = os.path.join(root_dir, 'dataset_' + phase)
    problem_path = os.path.join(root_dir, 'problem_' + phase)

    return dataset_path, problem_path


def _search_pipeline(dataset, problem, template, input_dir, output_dir,
                     budget, checkpoints, splits, db, tuner_type):

    dataset_path, problem_path = _get_dataset_paths(input_dir, dataset, 'TRAIN', problem)

    d3mds = D3MDS(dataset_path, problem_path)

    searcher = PipelineSearcher(
        output_dir,
        cv_splits=splits,
        db=db,
        tuner_type=tuner_type
    )

    return searcher.search(d3mds, template, budget=budget, checkpoints=checkpoints)


def _test_pipeline(dataset, problem, pipeline_id, input_dir, output_dir):

    dataset_path, problem_path = _get_dataset_paths(input_dir, dataset, 'TEST', problem)

    pipeline_path = os.path.join(output_dir, '{}.pkl'.format(pipeline_id))
    with open(pipeline_path, 'rb') as pipeline_pkl:
        pipeline = cloudpickle.load(pipeline_pkl)

    print('Executing best pipeline {}'.format(pipeline))

    d3mds = D3MDS(dataset_path, problem_path)

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        predictions = pipeline.predict(d3mds)

    return predictions


def _score_predictions(dataset, problem, predictions, input_dir):

    dataset_path, problem_path = _get_dataset_paths(input_dir, dataset, 'TEST', problem)
    metric = _get_metric(problem_path)

    targets = _load_targets(input_dir, dataset, problem)
    predictions = predictions.set_index('d3mIndex')[targets.columns]

    if len(targets.columns) > 1 or len(predictions.columns) > 1:
        raise Exception("I don't know how to handle these")

    if any(targets.index != predictions.index):
        raise Exception("Different indexes cannot be compared")

    targets = targets.iloc[:, 0]
    predictions = predictions.iloc[:, 0]
    score = metric(targets, predictions)
    print("Score: {}".format(score))

    summary = {'predictions': predictions, 'targets': targets}
    print(pd.DataFrame(summary).describe())

    return score


def _format_exception(e):
    error = '{}'.format(e.__class__.__name__)
    str_e = str(e)
    if str_e:
        error += ' - ' + str_e

    return error


def _score_dataset(dataset, args):

    start_ts = datetime.utcnow()

    result_base = {
        'dataset': dataset,
        'score': None,
        'elapsed': None,
        'iterations': None,
        'error': None,
        'step': None,
        'load_time': None,
        'trivial_time': None,
        'fit_time': None,
        'cv_time': None,
        'cv_score': None,
        'rank': None
    }
    results = []
    step = None
    try:
        step = 'SEARCH'
        print('###################')
        print('#### Searching ####')
        print('###################')

        # cleanup
        if not args.keep:
            shutil.rmtree(args.output, ignore_errors=True)

        search_results = _search_pipeline(
            dataset, args.problem, args.template, args.input, args.output, args.budget,
            args.checkpoints, args.splits, args.db, args.tuner_type
        )

        for search_result in search_results or []:
            result = result_base.copy()
            result.update(search_result)
            results.append(result)

            pipeline = result['pipeline']
            try:
                step = 'TEST'
                print('###################')
                print('#### Executing ####')
                print('###################')
                predictions = _test_pipeline(dataset, args.problem, pipeline,
                                             args.input, args.output)

                step = 'SCORE'
                print('#################')
                print('#### Scoring ####')
                print('#################')
                result['score'] = _score_predictions(dataset, args.problem,
                                                     predictions, args.input)

            except Exception as e:
                error = _format_exception(e)
                print("Scoring pipeline {} for dataset {} failed on step {} with error {}"
                      .format(pipeline, dataset, step, error))
                traceback.print_exc()
                result['error'] = error
                result['step'] = step

    except Exception as e:
        error = _format_exception(e)
        print("Dataset {} failed on step {} with error {}".format(dataset, step, error))
        traceback.print_exc()

        result_base['step'] = step
        result_base['error'] = error
        result_base['elapsed'] = (datetime.utcnow() - start_ts).total_seconds()
        results.append(result_base)

    return results


def _prepare_search(args):

    make_keras_picklable()

    if not args.datasets and not args.all:
        print('Please provide at least one dataset name or add the --all option')
        sys.exit(1)

    args.datasets = _get_datasets(args)

    if args.db:
        args.db = get_db(
            config=args.db_config,
            database=args.db_name,
            host=args.db_host,
            port=args.db_port,
            user=args.db_user,
            password=args.db_password
        )

    if args.checkpoints:
        args.checkpoints = [int(c) for c in args.checkpoints.split(',')]
    elif args.timeout:
        args.checkpoints = [args.timeout]


def _score_datasets(args):

    if args.report and os.path.exists(args.report):
        report = pd.read_csv(args.report)

    else:
        report = pd.DataFrame(columns=['dataset'])

    for dataset, row in args.datasets.iterrows():
        dataset_score = report[report.dataset == dataset]
        if dataset_score.empty or dataset_score.score.isnull().values[0]:

            if not dataset_score.empty:
                # clean-up
                report = report[report.dataset != dataset].copy()

            scores = _score_dataset(dataset, args)
            if scores:
                scores = pd.DataFrame(scores)
                scores = scores.merge(pd.DataFrame([row]), left_on='dataset', right_index=True)
                report = report.append(scores, ignore_index=True, sort=False)
                report = report.reindex(REPORT_COLUMNS, axis=1)

            if args.report:
                report.to_csv(args.report, index=False)

    return report


def _search(args):
    _prepare_search(args)

    print("Processing Datasets: {}".format(args.datasets.index.values))
    report = _score_datasets(args)

    report = report.reindex(REPORT_COLUMNS, axis=1)
    columns = REPORT_COLUMNS[1:]
    print(report.set_index('dataset').to_string(columns=columns))


def _get_datasets(args):
    if args.all:
        datasets = [
            d for d in os.listdir(args.input)
            if os.path.isdir(os.path.join(args.input, d))
        ]
    else:
        datasets = args.datasets

    exclude = getattr(args, 'exclude', None) or []
    datasets = [dataset for dataset in datasets if dataset not in exclude]

    try:
        summary = get_stats(datasets, args.input)
    except KeyError:
        print("No matching datasets found")
        sys.exit(1)

    summary = summary.set_index('dataset').reindex(datasets)
    summary = summary[~summary.data_modality.isnull()]

    for field in ['data_modality', 'task_type', 'task_subtype']:
        value = getattr(args, field)
        if value:
            summary = summary[summary[field] == value]

    if summary.empty:
        print("No matching datasets found")
        sys.exit(1)

    return summary


REPORT_COLUMNS = [
    'dataset',
    'pipeline',
    'score',
    'rank',
    'cv_score',
    'metric',
    'data_modality',
    'task_type',
    'task_subtype',
    'elapsed',
    'iterations',
    'load_time',
    'trivial_time',
    'fit_time',
    'cv_time',
    'error',
    'step'
]


def _list(args):
    args.all = True
    datasets = _get_datasets(args)
    datasets = datasets.reset_index().sort_values('dataset').set_index('dataset')
    columns = [
        'data_modality', 'task_type', 'task_subtype', 'metric', 'size_human', 'train_samples'
    ]
    datasets = datasets.reindex(columns, axis=1)
    print(datasets.to_string(columns=columns, index=True))


class ArgumentParser(argparse.ArgumentParser):
    def convert_arg_line_to_args(self, arg_line):
        return arg_line.split()


def _path_type(string):
    try:
        return make_abs(string)
    except ValueError:
        error = "Not a valid path: '{0}'.".format(string)
        raise argparse.ArgumentTypeError(error)


def _get_parser():

    # Logging
    logging_args = ArgumentParser(add_help=False)
    logging_args.add_argument('-v', '--verbose', action='count', default=0)
    logging_args.add_argument('-l', '--logfile')

    # Report
    report_args = ArgumentParser(add_help=False)
    report_args.add_argument('-r', '--report', type=_path_type,
                             help='Store results in the given CVS file.')

    # Dataset Selection
    dataset_args = ArgumentParser(add_help=False)
    dataset_args.add_argument('-i', '--input', default='data', type=_path_type,
                              help='Input datasets folder. Defaults to `data`.')
    dataset_args.add_argument('-o', '--output', type=_path_type,
                              help='Output pipelines folder. Defaults to `output`.',
                              default='output')
    dataset_args.add_argument('-p', '--problem', default='',
                              help='Problem suffix. Only needed if the dataset has more than one.')
    dataset_args.add_argument('-M', '--data-modality', type=str,
                              help='Dataset Modality.')
    dataset_args.add_argument('-T', '--task-type', type=str,
                              help='Dataset task type')
    dataset_args.add_argument('-S', '--task-subtype', type=str,
                              help='Dataset task subtype')

    # Search Configuration
    search_args = ArgumentParser(add_help=False)
    search_args.add_argument('-b', '--budget', type=int,
                             help=('Maximum number of tuning iterations to perform. '
                                   'Unlimited if not provided.'))
    search_args.add_argument('-s', '--splits', type=int, default=5,
                             help='Number of Cross Validation Folds. Defaults to 5')
    search_args.add_argument('-c', '--checkpoints',
                             help=('Comma separated list of time checkpoints where best pipeline '
                                   'so far will be dumped and stored in seconds, without spaces.'))
    search_args.add_argument('-t', '--timeout', type=int,
                             help='Timeout in seconds. Ignored if checkpoints are given.')
    search_args.add_argument('-u', '--tuner-type', default='gp', choices=TUNERS.keys(),
                             help='Type of tuner to use. Defaults to "gp"')
    search_args.add_argument('--template',
                             help='Template to use. If not given, use the most appropriate one.')
    search_args.add_argument('-e', '--exclude', nargs='+',
                             help='Exclude these datasets. Useful in combination with --all.')
    search_args.add_argument('-a', '--all', action='store_true',
                             help='Process all the datasets found in the input folder.')
    search_args.add_argument('-k', '--keep', action='store_true',
                             help='Keep previous results in the output folder.')
    search_args.add_argument('datasets', nargs='*',
                             help='Datasets to process. Ignored if --all use used.')

    # Backend configuration
    db_args = ArgumentParser(add_help=False)
    db_args.add_argument('--db', action='store_true',
                         help='Use a MongoDB backend to store the results.')
    db_args.add_argument('--db-config', help='MongoDB configuraiton JSON file.')
    db_args.add_argument('--db-host', default='localhost')
    db_args.add_argument('--db-port', default=27017, type=int)
    db_args.add_argument('--db-name', default='autobazaar')
    db_args.add_argument('--db-user')
    db_args.add_argument('--db-password')

    parser = ArgumentParser(
        description='AutoBazaar Experiments Suite',
        fromfile_prefix_chars='@',
        parents=[logging_args]
    )

    subparsers = parser.add_subparsers(title='command', help='Command to execute')
    parser.set_defaults(command=None)

    list_ = subparsers.add_parser('list', parents=[logging_args, dataset_args],
                                  help='List the available datasets that match the conditions.')
    list_.set_defaults(command=_list)

    search_parents = [
        logging_args,
        dataset_args,
        search_args,
        report_args,
        db_args
    ]
    search_ = subparsers.add_parser('search', parents=search_parents,
                                    help='Search the best pipeline for the given datasets.')
    search_.set_defaults(command=_search)

    return parser


def main():
    parser = _get_parser()
    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        parser.exit()

    logging_setup(args.verbose, args.logfile)

    args.command(args)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""
AutoBazaar top module.

AutoBazaar is an AutoML system created to execute the experiments associated with the
[The Machine Learning Bazaar Paper: Harnessing the ML Ecosystem for Effective System
Development](https://arxiv.org/pdf/1905.08942.pdf)
by the [Human-Data Interaction (HDI) Project](https://hdi-dai.lids.mit.edu/) at LIDS, MIT.

* Free software: MIT license
* Documentation: https://HDI-Project.github.io/MLBlocks
"""

__author__ = 'MIT Data To AI Lab'
__copyright__ = 'Copyright (c) 2019, MIT Data To AI Lab'
__email__ = 'dailabmit@gmail.com'
__license__ = 'MIT'
__version__ = '0.1.0'



# -*- coding: utf-8 -*-

"""AutoBazaar Search Module.


This module contains the PipelineSearcher, which is the class that
contains the main logic of the Auto Machine Learning process.
"""

import itertools
import json
import logging
import os
import signal
import warnings
from collections import defaultdict
from datetime import datetime

import numpy as np
from btb import HyperParameter
from btb.tuning import GP, GPEi, Uniform
from mit_d3m.loaders import get_loader
from mlblocks.mlpipeline import MLPipeline
from sklearn.model_selection import KFold, StratifiedKFold

from autobazaar.pipeline import ABPipeline
from autobazaar.utils import ensure_dir, make_dumpable, remove_dots, restore_dots

LOGGER = logging.getLogger(__name__)

warnings.filterwarnings("ignore", category=DeprecationWarning)

TEMPLATES_DIR = os.path.join(os.path.dirname(__file__), 'templates')


TRIVIAL_PIPELINE_METHOD = {
    'classification': 'mode',
    'regression': 'median',
    'collaborativeFiltering': 'median',
    'graphMatching': 'mode',
}

TUNERS = {
    'gp': GP,
    'gpei': GPEi,
    'uniform': Uniform
}

PARAM_TYPES = {
    'str': 'string',
}


class StopSearch(Exception):
    pass


class PipelineSearcher(object):
    """PipelineSearcher class.

    This class is responsible for searching the best pipeline to solve a
    given dataset and problem.
    """

    def __init__(self, pipelines_dir, db=None, test_id=None, tuner_type='gp',
                 cv_splits=5, random_state=0):

        self._db = db

        self._pipelines_dir = pipelines_dir
        ensure_dir(self._pipelines_dir)

        self._cv_splits = cv_splits
        self._random_state = random_state

        self._tuner_type = tuner_type
        self._tuner_class = TUNERS[tuner_type]

        self._test_id = test_id

    def _dump(self, pipeline):
        if not pipeline.dumped:
            pipeline.fit(self.data_params)

            mlpipeline = pipeline.pipeline

            LOGGER.info("Dumping pipeline %s: %s", pipeline.id, pipeline.pipeline)
            LOGGER.info("Hyperparameters: %s", mlpipeline.get_hyperparameters())
            pipeline.dump(self._pipelines_dir)

        else:
            LOGGER.info("Skipping already dumped pipeline %s", pipeline.id)

        self.pipelines.append({
            'elapsed': (datetime.utcnow() - self.start_time).total_seconds(),
            'iterations': len(self.solutions) - 1,
            'cv_score': self.best_pipeline.score,
            'rank': self.best_pipeline.rank,
            'pipeline': self.best_pipeline.id,
            'load_time': self.load_time,
            'trivial_time': self.trivial_time,
            'fit_time': self.fit_time,
            'cv_time': self.cv_time
        })

    def _save_pipeline(self, pipeline):
        solution = pipeline.to_dict(True)
        solution['_id'] = pipeline.id
        solution['ts'] = datetime.utcnow()

        self.solutions.append(solution)

        if self._db:
            insertable = remove_dots(solution)
            insertable.pop('problem_doc')
            insertable['dataset'] = self.dataset_id
            insertable['tuner_type'] = self._tuner_type
            insertable['test_id'] = self._test_id
            self._db.solutions.insert_one(insertable)

    def _build_trivial_pipeline(self):
        LOGGER.info("Building the Trivial pipeline")
        try:
            method = TRIVIAL_PIPELINE_METHOD.get(self.task_type)
            pipeline_dict = {
                'name': 'trivial.{}'.format(method),
                'primitives': ['mlprimitives.trivial.TrivialPredictor'],
                'init_params': {
                    'mlprimitives.trivial.TrivialPredictor': {
                        'method': method
                    }
                }
            }
            pipeline = ABPipeline(pipeline_dict, self.loader, self.metric, self.problem_doc)
            pipeline.cv_score(self.data_params.X, self.data_params.y,
                              self.data_params.context, cv=self.kf)

            self._save_pipeline(pipeline)

            return pipeline

        except Exception:
            # if the Trivial pipeline crashes we can do nothing,
            # so we just log the error and move on.
            LOGGER.exception("The Trivial pipeline crashed.")

    def _load_template_json(self, template_name):
        if template_name.endswith('.json'):
            template_filename = template_name
            name = template_name[:-5]
        else:
            name = template_name
            template_name = template_name.replace('/', '.') + '.json'
            template_filename = os.path.join(TEMPLATES_DIR, template_name)

        if os.path.exists(template_filename):
            with open(template_filename, 'r') as template_file:
                template_dict = json.load(template_file)
                template_dict['name'] = name

            return template_dict

    def _find_template(self, template_name):
        match = {
            'metadata.name': template_name
        }
        cursor = self._db.pipelines.find(match)
        templates = list(cursor.sort('metadata.insert_ts', -1).limit(1))
        if templates:
            template = templates[0]
            template['name'] = template.pop('metadata')['name']
            template['template'] = str(template.pop('_id'))
            return restore_dots(template)

    def _load_template(self, template_name):
        if self._db:
            template = self._find_template(template_name)
            if template:
                return template

        return self._load_template_json(template_name)

    def _get_template(self, template_name=None):
        if template_name:
            template = self._load_template(template_name)
            if not template:
                raise ValueError("Template {} not found".format(template_name))

            return template

        else:
            problem_type = [
                self.data_modality,
                self.task_type,
                self.task_subtype
            ]

            for levels in reversed(range(1, 4)):
                # Try the following options:
                # modality/task/subtask/default
                # modality/task/default
                # modality/default
                template_name = '/'.join(problem_type[:levels] + ['default'])
                template = self._load_template(template_name)
                if template:
                    return template

            # Nothing has been found for this modality/task/subtask combination
            raise StopSearch()

    def _build_default_pipeline(self, template_dict):
        LOGGER.info("Building the default pipeline")
        pipeline = ABPipeline(template_dict, self.loader, self.metric, self.problem_doc)

        X = self.data_params.X
        y = self.data_params.y
        context = self.data_params.context

        X = pipeline.preprocess(X, y, context)

        try:
            pipeline.cv_score(X, y, context, cv=self.kf)

            LOGGER.info("Saving the default pipeline %s", pipeline.id)

            self._save_pipeline(pipeline)
        except Exception:
            # if the Default pipeline crashes we can do nothing,
            # so we just log the error and move on.
            LOGGER.exception("The Default pipeline crashed.")
            pipeline = None

        return pipeline, X, y, context

    def _get_tuner(self, pipeline, template_dict):
        # Build an MLPipeline to get the tunables and the default params
        mlpipeline = MLPipeline.from_dict(template_dict)

        tunables = []
        tunable_keys = []
        for block_name, params in mlpipeline.get_tunable_hyperparameters().items():
            for param_name, param_details in params.items():
                key = (block_name, param_name)
                param_type = param_details['type']
                param_type = PARAM_TYPES.get(param_type, param_type)
                if param_type == 'bool':
                    param_range = [True, False]
                else:
                    param_range = param_details.get('range') or param_details.get('values')

                value = HyperParameter(param_type, param_range)
                tunables.append((key, value))
                tunable_keys.append(key)

        # Create the tuner
        LOGGER.info('Creating %s tuner', self._tuner_class.__name__)
        tuner = self._tuner_class(tunables)

        if pipeline:
            # Add the default params and the score obtained by the default pipeline to the tuner.
            default_params = defaultdict(dict)
            for block_name, params in pipeline.pipeline.get_hyperparameters().items():
                for param, value in params.items():
                    key = (block_name, param)
                    if key in tunable_keys:
                        # default_params[key] = 'None' if value is None else value
                        default_params[key] = value

            tuner.add(default_params, 1 - pipeline.rank)

        return tuner

    def _set_checkpoint(self):
        next_checkpoint = self.checkpoints.pop(0)
        interval = next_checkpoint - self.current_checkpoint
        LOGGER.info("Setting %s seconds checkpoint in %s seconds", next_checkpoint, interval)
        signal.alarm(interval)
        self.current_checkpoint = next_checkpoint

    def _checkpoint(self, signum=None, frame=None, final=False):
        signal.alarm(0)

        checkpoint_name = 'Final' if final else str(self.current_checkpoint) + ' seconds'

        LOGGER.info("%s checkpoint reached", checkpoint_name)

        set_checkpoint = (not final) and bool(self.checkpoints)
        if set_checkpoint:
            self._set_checkpoint()

        try:
            if self.best_pipeline:
                self._dump(self.best_pipeline)

        except Exception:
            LOGGER.exception("Checkpoint dump crashed")

        if not set_checkpoint:
            self.current_checkpoint = None
            raise StopSearch()

    def search(self, d3mds, template_name=None, budget=None, checkpoints=None):
        # Problem variables
        problem_id = d3mds.get_problem_id()
        self.task_type = d3mds.get_task_type()
        self.task_subtype = d3mds.problem.get_task_subtype()

        if self.task_type == 'classification':
            self.kf = StratifiedKFold(
                n_splits=self._cv_splits,
                shuffle=True,
                random_state=self._random_state
            )
        else:
            self.kf = KFold(
                n_splits=self._cv_splits,
                shuffle=True,
                random_state=self._random_state
            )

        self.problem_doc = d3mds.problem_doc

        # Dataset variables
        self.dataset_id = d3mds.dataset_id
        self.data_modality = d3mds.get_data_modality()

        self.metric = d3mds.get_metric()

        self.loader = get_loader(self.data_modality, self.task_type)

        self.best_pipeline = None

        self.solutions = []
        self.checkpoints = sorted(checkpoints or [])
        self.current_checkpoint = 0
        self.pipelines = []

        if not self.checkpoints and budget is None:
            budget = 1

        self.load_time = None
        self.trivial_time = None
        self.fit_time = None
        self.cv_times = []
        self.cv_time = None

        try:
            self.start_time = datetime.utcnow()

            LOGGER.info("Running TA2 Search")
            LOGGER.info("Problem Id: %s", problem_id)
            LOGGER.info("    Data Modality: %s", self.data_modality)
            LOGGER.info("    Task type: %s", self.task_type)
            LOGGER.info("    Task subtype: %s", self.task_subtype)
            LOGGER.info("    Metric: %s", self.metric)
            LOGGER.info("    Checkpoints: %s", self.checkpoints)
            LOGGER.info("    Budget: %s", budget)

            if self.checkpoints:
                signal.signal(signal.SIGALRM, self._checkpoint)
                self._set_checkpoint()

            load_start = datetime.utcnow()
            self.data_params = self.loader.load(d3mds)
            load_end = datetime.utcnow()
            self.load_time = (load_end - load_start).total_seconds()

            # Build the trivial pipeline
            trivial_start = datetime.utcnow()
            self.best_pipeline = self._build_trivial_pipeline()
            trivial_end = datetime.utcnow()
            self.trivial_time = (trivial_end - trivial_start).total_seconds()

            template_dict = self._get_template(template_name)

            # Do not continue if there is no budget or no fit data
            if budget == 0 or not len(self.data_params.X):
                raise StopSearch()

            # Build the default pipeline
            fit_start = datetime.utcnow()
            default_pipeline, X, y, context = self._build_default_pipeline(template_dict)
            if default_pipeline:
                self.best_pipeline = default_pipeline

            fit_end = datetime.utcnow()
            self.fit_time = (fit_end - fit_start).total_seconds()

            if budget == 1:
                raise StopSearch()
            elif budget is not None:
                budget -= 1

            # Build the tuner
            tuner = self._get_tuner(default_pipeline, template_dict)

            LOGGER.info("Starting the tuning loop")

            if budget is not None:
                iterator = range(budget)
            else:
                iterator = itertools.count()   # infinite range

            for iteration in iterator:
                proposed_params = tuner.propose()
                params = make_dumpable(proposed_params)

                cv_start = datetime.utcnow()
                pipeline_dict = template_dict.copy()
                pipeline_dict['hyperparameters'] = params
                pipeline = ABPipeline(pipeline_dict, self.loader,
                                      self.metric, self.problem_doc)
                try:
                    pipeline.cv_score(X, y, context, cv=self.kf)
                except StopSearch:
                    raise

                except Exception:
                    LOGGER.exception("Crash during cross validation")
                    tuner.add(proposed_params, -1000000)

                if pipeline.rank is not None:
                    tuner.add(proposed_params, 1 - pipeline.rank)

                    cv_end = datetime.utcnow()
                    cv_time = (cv_end - cv_start).total_seconds()
                    self.cv_times.append(cv_time)
                    self.cv_time = np.mean(self.cv_times)

                    LOGGER.info("Saving pipeline %s: %s", iteration + 1, pipeline.id)
                    self._save_pipeline(pipeline)

                    if not self.best_pipeline or (pipeline.rank < self.best_pipeline.rank):
                        self.best_pipeline = pipeline
                        LOGGER.info('Best pipeline so far: %s; rank: %s, score: %s',
                                    self.best_pipeline, self.best_pipeline.rank,
                                    self.best_pipeline.score)

        except (StopSearch, KeyboardInterrupt):
            pass

        finally:
            signal.alarm(0)

        if self.current_checkpoint:
            self._checkpoint(final=True)
        elif self.best_pipeline and not checkpoints:
            self._dump(self.best_pipeline)

        if self.best_pipeline:
            LOGGER.info('Best pipeline for problem %s found: %s; rank: %s, score: %s',
                        problem_id, self.best_pipeline,
                        self.best_pipeline.rank, self.best_pipeline.score)

        else:
            LOGGER.info('No pipeline could be found for problem %s', problem_id)

        return self.pipelines



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md') as history_file:
    history = history_file.read()

install_requires = [
    'absl-py==0.4.0',
    'astor==0.7.1',
    'baytune==0.2.1',
    'boto==2.48.0',
    'boto3==1.9.27',
    'botocore==1.12.28',
    'certifi==2018.8.13',
    'chardet==3.0.4',
    'click==6.7',
    'cloudpickle==0.4.0',
    'cycler==0.10.0',
    'dask==0.18.2',
    'decorator==4.3.0',
    'distributed==1.22.1',
    'docutils==0.14',
    'featuretools==0.3.1',
    'future==0.16.0',
    'gast==0.2.0',
    'grpcio==1.12.1',
    'h5py==2.8.0',
    'HeapDict==1.0.0',
    'idna==2.6',
    'iso639==0.1.4',
    'jmespath==0.9.3',
    'Keras==2.1.6',
    'Keras-Applications==1.0.6',
    'Keras-Preprocessing==1.0.5',
    'kiwisolver==1.0.1',
    'langdetect==1.0.7',
    'lightfm==1.15',
    'matplotlib==2.2.3',
    'mit-d3m==0.1.1',
    'mlblocks==0.2.3',
    'mlprimitives==0.1.3',
    'msgpack==0.5.6',
    'networkx==2.1',
    'nltk==3.3',
    'numpy==1.15.2',
    'opencv-python==3.4.2.17',
    'pandas==0.23.4',
    'Pillow==5.1.0',
    'protobuf==3.6.1',
    'psutil==5.4.7',
    'pymongo==3.7.2',
    'pyparsing==2.2.0',
    'python-dateutil==2.7.3',
    'python-louvain==0.10',
    'pytz==2018.5',
    'PyWavelets==0.5.2',
    'PyYAML==3.12',
    'requests==2.20.0',
    's3fs==0.1.5',
    's3transfer==0.1.13',
    'scikit-image==0.14.0',
    'scikit-learn==0.20.0',
    'scipy==1.1.0',
    'six==1.11.0',
    'sortedcontainers==2.0.4',
    'setuptools==39.1.0',
    'tblib==1.3.2',
    'tensorboard==1.11.0',
    'tensorflow==1.11.0',
    'termcolor==1.1.0',
    'toolz==0.9.0',
    'tornado==5.1',
    'tqdm==4.24.0',
    'urllib3==1.23',
    'Werkzeug==0.14.1',
    'xgboost==0.72.1',
    'zict==0.1.3',
]

tests_require = [
    'pytest>=3.4.2',
    'pytest-cov>=2.6.0',
]

setup_requires = [
    'pytest-runner>=2.11.1',
]

development_requires = [
    # general
    'bumpversion>=0.5.3',
    'pip>=10.0.0',
    'watchdog>=0.8.3',

    # docs
    'm2r>=0.2.0',
    'Sphinx>=1.7.1',
    'sphinx_rtd_theme>=0.2.4',
    'autodocsumm>=0.1.10',

    # style check
    'flake8>=3.5.0',
    'isort>=4.3.4',

    # fix style issues
    'autoflake>=1.1',
    'autopep8>=1.3.5',

    # distribute on PyPI
    'twine>=1.10.0',
    'wheel>=0.30.0',

    # Advanced testing
    'tox>=2.9.1',
    'coverage>=4.5.1',
]

setup(
    author='MIT Data To AI Lab',
    author_email='dailabmit@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    description='The Machine Learning Bazaar',
    entry_points={
        'console_scripts': [
            'abz=autobazaar.__main__:main'
        ]
    },
    extras_require={
        'dev': development_requires + tests_require,
        'tests': tests_require,
    },
    include_package_data=True,
    install_requires=install_requires,
    license='MIT license',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    keywords='automl machine learning hyperparameters tuning classification regression autobazaar',
    name='autobazaar',
    packages=find_packages(include=['autobazaar', 'autobazaar.*']),
    python_requires='>=3.4',
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://github.com/HDI-project/AutoBazaar',
    version='0.1.0',
    zip_safe=False,
)
